from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...types.tool import ToolDef

class Toolset(ABC):
    def format_tool_name(self, tool_name: str) -> str:
        return f"{self.name}__{tool_name}"

    @property
    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    def get_tools(self) -> list[ToolDef]: ...



import dataclasses
import json
from ..types.event import TextChunkEvent, ToolCallChunkEvent, UsageChunkEvent
from ..types.message import AssistantMessage


class ToolCallCollector:
    @dataclasses.dataclass
    class ToolCallTemp:
        id: str = ""
        name: str = ""
        arguments: str = ""

    def __init__(self):
        self.tool_call_map: dict[int, ToolCallCollector.ToolCallTemp] = {}

    def collect(self, tool_call_chunk: ToolCallChunkEvent):
        if tool_call_chunk.index not in self.tool_call_map:
            self.tool_call_map[tool_call_chunk.index] = ToolCallCollector.ToolCallTemp()

        temp_tool_call = self.tool_call_map[tool_call_chunk.index]
        if tool_call_chunk.id:
            temp_tool_call.id = tool_call_chunk.id
        if tool_call_chunk.name:
            temp_tool_call.name += tool_call_chunk.name
        if tool_call_chunk.arguments:
            temp_tool_call.arguments += tool_call_chunk.arguments

    def get_tool_calls(self) -> list[AssistantMessage.ToolCall]:
        return [AssistantMessage.ToolCall(
            id=tool_call.id,
            name=tool_call.name,
            arguments=json.loads(tool_call.arguments),
        ) for tool_call in self.tool_call_map.values()]

class StreamMessageCollector:
    def __init__(self):
        self._message_buf = AssistantMessage(content=None)
        self._tool_call_collector = ToolCallCollector()

    def collect(self, chunk: TextChunkEvent | ToolCallChunkEvent | UsageChunkEvent):
        match chunk:
            case ToolCallChunkEvent():
                self._tool_call_collector.collect(chunk)
            case TextChunkEvent():
                if self._message_buf.content is None:
                    self._message_buf.content = ""
                self._message_buf.content += chunk.content
            case UsageChunkEvent():
                if self._message_buf.usage is None:
                    self._message_buf.usage = AssistantMessage.Usage.default()
                # since the usage chunk is the last chunk, we can directly assign the values here
                self._message_buf.usage.input_tokens = chunk.input_tokens
                self._message_buf.usage.output_tokens = chunk.output_tokens
                self._message_buf.usage.total_tokens = chunk.total_tokens

    def get_message(self) -> AssistantMessage:
        self._message_buf.tool_calls = self._tool_call_collector.get_tool_calls()
        # if self._message_buf.content is not None and self._message_buf.reasoning_content is None:
        #     self._message_buf.content, self._message_buf.reasoning_content =\
        #         AssistantMessage.extract_thinking_content(self._message_buf.content)
        result = self._message_buf

        # reset the message buffer
        self._message_buf = AssistantMessage(content=None)
        self._tool_call_collector = ToolCallCollector()
        return result



from dataclasses import dataclass, field
from typing import Any, Literal, Sequence, TYPE_CHECKING

if TYPE_CHECKING:
    from .tool import ToolLike
    from ..tool.toolset import Toolset
    from .message import ChatMessage

@dataclass
class LlmRequestParams:
    model: str
    messages: Sequence[ChatMessage]
    instructions: str | None = None
    tools: list[ToolLike] | None = None
    toolsets: list[Toolset] | None = None
    tool_choice: Literal["auto", "required", "none"] = "auto"

    timeout_sec: float | None = None
    temperature: float | None = None
    max_tokens: int | None = None
    headers: dict[str, str] | None = None

    extra_args: dict[str, Any] | None = None

    _extract_tools_cache: list[ToolLike] | None = field(default=None, init=False, repr=False)

    def extract_tools(self) -> list[ToolLike] | None:
        if self._extract_tools_cache is not None:
            return self._extract_tools_cache

        if self.tools is None and self.toolsets is None:
            return None
        tools = []
        if self.tools:
            tools = self.tools
        if self.toolsets:
            for toolset in self.toolsets:
                tools.extend(toolset.get_tools())

        self._extract_tools_cache = tools
        return tools

    def find_tool(self, tool_name: str) -> ToolLike | None:
        from ..tool.utils import find_tool_by_name

        has_tool = ((self.tools is not None and len(self.tools) > 0) or
                        (self.toolsets is not None and len(self.toolsets) > 0))
        if not has_tool: return None

        if (tools := self.extract_tools()) is None:
            return None
        return find_tool_by_name(tools, tool_name)

__all__ = [
    "LlmRequestParams",
]



"""
source: https://github.com/mozilla-ai/any-llm/blob/main/src/any_llm/tools.py
"""

import dataclasses
import enum
import inspect
import types as _types
from collections.abc import Mapping, Sequence
from datetime import date, datetime, time
from typing import (Annotated as _Annotated, Literal as _Literal,
                    is_typeddict as _is_typeddict, Any, get_args,
                    get_origin, get_type_hints)
from pydantic import BaseModel as PydanticBaseModel
from ..types.tool import ToolFn, ToolDef, RawToolDef, ToolLike, ToolSchema

def _python_type_to_json_schema(python_type: Any) -> dict[str, Any]:
    """Convert Python type annotation to a JSON Schema for a parameter.

    Supported mappings (subset tailored for LLM tool schemas):
    - Primitives: str/int/float/bool -> string/integer/number/boolean
    - bytes -> string with contentEncoding base64
    - datetime/date/time -> string with format date-time/date/time
    - list[T] / Sequence[T] / set[T] / frozenset[T] -> array with items=schema(T)
      - set/frozenset include uniqueItems=true
      - list without type args defaults items to string
    - dict[K,V] / Mapping[K,V] -> object with additionalProperties=schema(V)
      - dict without type args defaults additionalProperties to string
    - tuple[T1, T2, ...] -> array with prefixItems per element and min/maxItems
    - tuple[T, ...] -> array with items=schema(T)
    - Union[X, Y] and X | Y -> oneOf=[schema(X), schema(Y)] (without top-level type)
    - Optional[T] (Union[T, None]) -> oneOf=[schema(T), {"type": "null"}]
    - Literal[...]/Enum -> enum with appropriate type inference when uniform
    - TypedDict -> object with properties/required per annotations
    - dataclass/Pydantic BaseModel -> object with nested properties inferred from fields
    """
    origin = get_origin(python_type)
    args = get_args(python_type)

    if _Annotated is not None and origin is _Annotated and len(args) >= 1:
        python_type = args[0]
        origin = get_origin(python_type)
        args = get_args(python_type)

    if python_type is Any:
        return {"type": "string"}

    primitive_map = {str: "string", int: "integer", float: "number", bool: "boolean"}
    if python_type in primitive_map:
        return {"type": primitive_map[python_type]}

    if python_type is bytes:
        return {"type": "string", "contentEncoding": "base64"}
    if python_type is datetime:
        return {"type": "string", "format": "date-time"}
    if python_type is date:
        return {"type": "string", "format": "date"}
    if python_type is time:
        return {"type": "string", "format": "time"}

    if python_type is list:
        return {"type": "array", "items": {"type": "string"}}
    if python_type is dict:
        return {"type": "object", "additionalProperties": {"type": "string"}}

    if origin is _Literal:
        literal_values = list(args)
        schema_lit: dict[str, Any] = {"enum": literal_values}
        if all(isinstance(v, bool) for v in literal_values):
            schema_lit["type"] = "boolean"
        elif all(isinstance(v, str) for v in literal_values):
            schema_lit["type"] = "string"
        elif all(isinstance(v, int) and not isinstance(v, bool) for v in literal_values):
            schema_lit["type"] = "integer"
        elif all(isinstance(v, int | float) and not isinstance(v, bool) for v in literal_values):
            schema_lit["type"] = "number"
        return schema_lit

    if inspect.isclass(python_type) and issubclass(python_type, enum.Enum):
        enum_values = [e.value for e in python_type]
        value_types = {type(v) for v in enum_values}
        schema: dict[str, Any] = {"enum": enum_values}
        if value_types == {str}:
            schema["type"] = "string"
        elif value_types == {int}:
            schema["type"] = "integer"
        elif value_types <= {int, float}:
            schema["type"] = "number"
        elif value_types == {bool}:
            schema["type"] = "boolean"
        return schema

    if _is_typeddict(python_type):
        annotations: dict[str, Any] = getattr(python_type, "__annotations__", {}) or {}
        required_keys = set(getattr(python_type, "__required_keys__", set()))
        td_properties: dict[str, Any] = {}
        td_required: list[str] = []
        for field_name, field_type in annotations.items():
            td_properties[field_name] = _python_type_to_json_schema(field_type)
            if field_name in required_keys:
                td_required.append(field_name)
        schema_td: dict[str, Any] = {
            "type": "object",
            "properties": td_properties,
        }
        if td_required:
            schema_td["required"] = td_required
        return schema_td

    if inspect.isclass(python_type) and dataclasses.is_dataclass(python_type):
        type_hints = get_type_hints(python_type)
        dc_properties: dict[str, Any] = {}
        dc_required: list[str] = []
        for field in dataclasses.fields(python_type):
            field_type = type_hints.get(field.name, Any)
            dc_properties[field.name] = _python_type_to_json_schema(field_type)
            if (
                field.default is dataclasses.MISSING
                and getattr(field, "default_factory", dataclasses.MISSING) is dataclasses.MISSING
            ):
                dc_required.append(field.name)
        schema_dc: dict[str, Any] = {"type": "object", "properties": dc_properties}
        if dc_required:
            schema_dc["required"] = dc_required
        return schema_dc

    if inspect.isclass(python_type) and issubclass(python_type, PydanticBaseModel):
        model_type_hints = get_type_hints(python_type)
        pd_properties: dict[str, Any] = {}
        pd_required: list[str] = []
        model_fields = getattr(python_type, "model_fields", {})
        for name, field_info in model_fields.items():
            pd_properties[name] = _python_type_to_json_schema(model_type_hints.get(name, Any))
            is_required = getattr(field_info, "is_required", None)
            if callable(is_required) and is_required():
                pd_required.append(name)
        schema_pd: dict[str, Any] = {"type": "object", "properties": pd_properties}
        if pd_required:
            schema_pd["required"] = pd_required
        return schema_pd

    if origin in (list, Sequence, set, frozenset):
        item_type = args[0] if args else Any
        item_schema = _python_type_to_json_schema(item_type)
        schema_arr: dict[str, Any] = {"type": "array", "items": item_schema or {"type": "string"}}
        if origin in (set, frozenset):
            schema_arr["uniqueItems"] = True
        return schema_arr
    if origin is tuple:
        if not args:
            return {"type": "array", "items": {"type": "string"}}
        if len(args) == 2 and args[1] is Ellipsis:
            return {"type": "array", "items": _python_type_to_json_schema(args[0])}
        prefix_items = [_python_type_to_json_schema(a) for a in args]
        return {
            "type": "array",
            "prefixItems": prefix_items,
            "minItems": len(prefix_items),
            "maxItems": len(prefix_items),
        }

    if origin in (dict, Mapping):
        value_type = args[1] if len(args) >= 2 else Any
        value_schema = _python_type_to_json_schema(value_type)
        return {"type": "object", "additionalProperties": value_schema or {"type": "string"}}

    typing_union = getattr(__import__("typing"), "Union", None)
    if origin in (typing_union, _types.UnionType):
        non_none_args = [a for a in args if a is not type(None)]
        has_none = len(non_none_args) < len(args)

        if len(non_none_args) > 1:
            schemas = [_python_type_to_json_schema(arg) for arg in non_none_args]
            if has_none:
                schemas.append({"type": "null"})
            return {"oneOf": schemas}

        if non_none_args:
            inner = _python_type_to_json_schema(non_none_args[0])
            if has_none:
                # if inner is `oneOf`, append null to it and prevent nested oneOf
                if "oneOf" in inner:
                    inner["oneOf"].append({"type": "null"})
                    return inner
                return {"oneOf": [inner, {"type": "null"}]}
            return inner

        return {"type": "null"} # for `Union[None]`

    return {"type": "string"}

def _parse_description(python_type: Any) -> str | None:
    origin = get_origin(python_type)
    if origin is not _Annotated: return None

    args = get_args(python_type)
    metadata = args[1:]

    if len(metadata) == 0: return None
    if not isinstance(metadata[0], str): return None
    return inspect.cleandoc(metadata[0])

def _parse_callable_properties(func: ToolFn) -> tuple[dict[str, dict[str, Any]], list[str]]:
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)
    annotated_type_hints = get_type_hints(func, include_extras=True)

    properties: dict[str, dict[str, Any]] = {}
    required: list[str] = []

    for param_name, param in sig.parameters.items():
        if param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
            continue

        type_hint = type_hints.get(param_name, str)
        param_schema = _python_type_to_json_schema(type_hint)

        annotated_type = annotated_type_hints.get(param_name)
        param_description = _parse_description(annotated_type)

        type_name = getattr(type_hint, "__name__", str(type_hint))
        properties[param_name] = {
            **param_schema,
            "description": param_description or f"Parameter {param_name} of type {type_name}",
        }

        if param.default == inspect.Parameter.empty:
            required.append(param_name)

    return properties, required

def generate_tool_definition_from_callable(func: ToolFn) -> ToolSchema:
    """Convert a Python callable to OpenAI tools format.

    Args:
        func: A Python callable (function) to convert to a tool

    Returns:
        Dictionary in OpenAI tools format

    Raises:
        ValueError: If the function doesn't have proper docstring or type annotations

    Example:
        ```
        >>> def get_weather(location: str, unit: str = "celsius") -> str:
        ...     '''Get weather information for a location.'''
        ...     return f"Weather in {location} is sunny, 25°{unit[0].upper()}"
        >>>
        >>> tool = generate_tool_definition_from_callable(get_weather)
        >>> # Returns OpenAI tools format dict
        ```

    """
    if not func.__doc__:
        msg = f"Function {func.__name__} must have a docstring"
        raise ValueError(msg)

    properties, required = _parse_callable_properties(func)
    return ToolSchema(
        name=func.__name__,
        description=inspect.cleandoc(func.__doc__),
        parameters={"type": "object", "properties": properties, "required": required},
    )

def generate_tool_definition_from_tool_def(tool_def: ToolDef) -> ToolSchema:
    """Convert a ToolDef to OpenAI tools format.

    Args:
        tool_def: A ToolDef to convert to a tool

    Returns:
        Dictionary in OpenAI tools format

    Example:
        >>> tool_def = ToolDef(
        ...     name="get_weather",
        ...     description="Get weather information for a location.",
        ...     execute=SomeFunction(),
        ... )
        >>> tool = generate_tool_definition_from_tool_def(tool_def)
        >>> # Returns OpenAI tools format dict
    """
    properties, required = _parse_callable_properties(tool_def.execute)
    return ToolSchema(
        name=tool_def.name,
        description=tool_def.description,
        parameters=tool_def.parameters or {"type": "object", "properties": properties, "required": required},
    )

def generate_tool_definition_from_raw_tool_def(raw_tool_def: RawToolDef) -> ToolSchema:
    return raw_tool_def

def prepare_tools(tools: Sequence[ToolLike]) -> list[ToolSchema]:
    tool_likes = []
    for tool in tools:
        if callable(tool):
            tool_likes.append(generate_tool_definition_from_callable(tool))
        elif isinstance(tool, ToolDef):
            tool_likes.append(generate_tool_definition_from_tool_def(tool))
        else:
            tool_likes.append(generate_tool_definition_from_raw_tool_def(tool))
    return tool_likes



import json
import inspect
from functools import singledispatch
from typing import Any, Callable, assert_never, cast
from types import FunctionType, MethodType
from ..types.tool import ToolDef, ToolLike
from ..types.exceptions import LlmToolException
from ..logger import logger

type ExceptionHandler[E: LlmToolException] = Callable[[E], str]

class ToolExceptionHandlerManager:
    def __init__(self):
        self._handlers: dict[type[LlmToolException], ExceptionHandler[Any]] = {}

    def register[E: LlmToolException](self, exception_type: type[E]):
        def decorator(handler: ExceptionHandler[E]) -> ExceptionHandler[E]:
            self.set_handler(exception_type, handler)
            return handler
        return decorator

    def set_handler[E: LlmToolException](self, exception_type: type[E], handler: ExceptionHandler[E]):
        self._handlers[exception_type] = handler

    def get_handler[E: LlmToolException](self, exception_type: type[E]) -> ExceptionHandler[E] | None:
        return self._handlers.get(exception_type)

    def handle(self, e: LlmToolException) -> str:
        def find_best_handler[E: LlmToolException](exc_type: type[E]) -> ExceptionHandler[E] | None:
            for cls in exc_type.__mro__:
                if cls in self._handlers:
                    return self._handlers[cls]
            return None

        # Searches the MRO of the exception type to make sure the subclasses of
        # the registered exception type can also be handled.
        handler = find_best_handler(type(e))
        if handler is None:
            logger.warning(f"Unhandled tool exception: {type(e).__name__}", exc_info=e)
            return f"Unhandled tool exception | {type(e).__name__}: {e}"
        return handler(e)

# --- --- --- --- --- ---

def _arguments_normalizer(arguments: str | dict) -> dict:
    if isinstance(arguments, str):
        if len(arguments.strip()) == 0:
            return {}
        parsed = json.loads(arguments)
        return cast(dict, parsed)
    elif isinstance(arguments, dict):
        return arguments
    else:
        assert_never(arguments)

def _result_normalizer(result: Any) -> str:
    if isinstance(result, str):
        return result
    return json.dumps(result, ensure_ascii=False)

@singledispatch
async def execute_tool(tool: ToolLike, arguments: str | dict) -> str:
    """
    Raises:
        ValueError: If the tool type is not supported.
        JSONDecodeError: If the arguments is a string but not valid JSON.
    """
    raise ValueError(f"Invalid tool type: {type(tool)}")

@execute_tool.register(FunctionType)
@execute_tool.register(MethodType)
async def _(toolfn: Callable, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = (await toolfn(**arguments)
             if inspect.iscoroutinefunction(toolfn)
             else toolfn(**arguments))
    return _result_normalizer(result)

@execute_tool.register(ToolDef)
async def _(tooldef: ToolDef, arguments: str | dict) -> str:
    arguments = _arguments_normalizer(arguments)
    result = (await tooldef.execute(**arguments)
             if inspect.iscoroutinefunction(tooldef.execute)
             else tooldef.execute(**arguments))
    return _result_normalizer(result)



import json
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .tool import ToolLike

class LlmToolException(Exception): ...

class ToolDoesNotExistError(LlmToolException):
    def __init__(self, tool_name: str):
        super().__init__(f"Tool does not exist: ", tool_name)
        self.tool_name = tool_name

class ToolArgumentDecodeError(LlmToolException):
    def __init__(self, tool_name: str, arguments: str, raw_error: json.JSONDecodeError):
        super().__init__(f"Tool argument decode error: ", raw_error)
        self.tool_name = tool_name
        self.arguments = arguments
        self.raw_error = raw_error

class ToolExecutionError(LlmToolException):
    def __init__(self, tool: ToolLike, arguments: str | dict, raw_error: Exception):
        super().__init__(f"Tool execution error: ", raw_error)
        self.tool = tool
        self.arguments = arguments
        self.raw_error = raw_error

class LlmProviderException(Exception): ...

class AttachmentTypeNotSupportedError(LlmProviderException):
    def __init__(self, attachment_type: str):
        super().__init__(f"Attachment type not supported: ", attachment_type)
        self.attachment_type = attachment_type

__all__ = [
    "LlmToolException",
    "ToolDoesNotExistError",
    "ToolArgumentDecodeError",
    "ToolExecutionError",

    "LlmProviderException",
    "AttachmentTypeNotSupportedError",
]



from typing import assert_never
from ..types.tool import ToolDef, ToolLike

def get_tool_name(tool: ToolLike) -> str:
    if callable(tool):
        return tool.__name__
    elif isinstance(tool, ToolDef):
        return tool.name
    elif isinstance(tool, dict):
        return tool.get("name", "")
    else:
        assert_never(tool)

def find_tool_by_name(tools: list[ToolLike], name: str) -> ToolLike | None:
    for tool in tools:
        if callable(tool) and tool.__name__ == name:
            return tool
        elif isinstance(tool, ToolDef) and tool.name == name:
            return tool
        elif isinstance(tool, dict) and tool.get("name") == name:
            return tool
    return None



import inspect
from typing import Any, Callable, override, overload
from collections.abc import Mapping
from pydantic import validate_call, ConfigDict
from .toolset import Toolset
from ...types.tool import ToolDef

TOOL_FLAG = "__is_tool__"
TOOL_DEFAULTS = "__tool_defaults__"

def is_tool(func: Callable) -> bool:
    return getattr(func, TOOL_FLAG, False)

def get_tool_defaults(func: Callable) -> Mapping[str, Any]:
    return getattr(func, TOOL_DEFAULTS, {})

@overload
def python_tool[F: Callable[..., Any]](func: F) -> F: ...

@overload
def python_tool[F: Callable[..., Any]](func: None = None,
                                       *,
                                       validate: bool = False,
                                       validate_config: ConfigDict | None = None,
                                       defaults: Mapping[str, Any] | None = None,
                                       ) -> Callable[[F], F]: ...

def python_tool[F: Callable[..., Any]](func: F | None = None,
                                       *,
                                       validate: bool = False,
                                       validate_config: ConfigDict | None = None,
                                       defaults: Mapping[str, Any] | None = None,
                                       ) -> F | Callable[[F], F]:
    """
    Mark a callable as a tool and optionally enable runtime argument validation.

    This decorator supports both forms:
    - @python_tool
    - @python_tool(validate=True, validate_config=...)

    When validation is enabled, the target callable is wrapped by `pydantic.validate_call`,
    so arguments are validated (and may be coerced) according to type annotations before
    execution. The internal tool flag is attached to the final callable so it can be
    recognized by ``PythonToolset.get_tools``.

    Args:
        func: The target callable when used as ``@python_tool``. ``None`` when used as
            a configurable decorator factory.
        validate: Whether to enable Pydantic runtime argument validation.
            Defaults to ``False``.
        validate_config: Optional Pydantic ``ConfigDict`` passed to
            ``validate_call(config=...)``.
        defaults: Optional static configs for tool.

    Returns:
        The decorated callable (when ``func`` is provided), or a decorator that accepts
        a callable and returns the decorated callable.

    Examples:
        ```python
        @python_tool
        def add(x: int, y: int) -> int:
            return x + y

        @python_tool(validate=True)
        def add_checked(x: int, y: int) -> int:
            return x + y

        @python_tool(defaults={"auto_approve": True})
        def read_file(path: str) -> str:
            ...
        ```
    """

    def decorator(f: F) -> F:
        if validate:
            f = validate_call(config=validate_config)(f)
        setattr(f, TOOL_FLAG, True)
        setattr(f, TOOL_DEFAULTS, dict(defaults or {}))
        return f

    if func is not None:
        return decorator(func)
    return decorator

# --- --- --- --- --- ---

class PythonToolset(Toolset):
    @property
    @override
    def name(self) -> str:
        """
        Since the usage of PythonToolset is to inherit and define methods as tools,
        the name of the toolset is the name of the subclass.
        """
        return self.__class__.__name__

    @override
    def get_tools(self, namespaced_tool_name: bool = True) -> list[ToolDef]:
        result = []
        for _, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not is_tool(method): continue
            tool_def = ToolDef.from_tool_fn(method)
            tool_def.name = (self.format_tool_name(tool_def.name)
                             if namespaced_tool_name
                             else tool_def.name)
            tool_def.defaults = get_tool_defaults(method)
            result.append(tool_def)
        return result



from .base_mcp_client import McpClient, McpTool, ToolResult as McpToolResult
from .local_mcp_client import LocalMcpClient, LocalServerParams
from .remote_mcp_client import RemoteMcpClient, RemoteServerParams, OAuthParams

__all__ = [
    "McpClient",
    "McpTool",
    "McpToolResult",

    "LocalMcpClient",
    "LocalServerParams",
    "RemoteMcpClient",
    "RemoteServerParams",
    "OAuthParams",
]



import json
import re
from typing import Literal, cast, override
from openai import AsyncOpenAI
from openai.types.shared_params import FunctionDefinition
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionFunctionToolParam,
    ChatCompletionMessageFunctionToolCallParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
    ChatCompletionAssistantMessageParam,
    ChatCompletionToolMessageParam,
)
from openai.types.chat.chat_completion_chunk import ChoiceDelta
from openai.types.chat.completion_create_params import CompletionCreateParamsNonStreaming, CompletionCreateParamsStreaming
from .base_provider import BaseProvider, BaseMessageParser, BaseParamParser
from .utils import StreamMessageCollector
from ..tool.prepare import prepare_tools
from ..types.attachment import Attachment, AudioAttachment, ImageAttachment
from ..types.exceptions import AttachmentTypeNotSupportedError
from ..types.request_params import LlmRequestParams
from ..types.message import ChatMessage, SystemMessage, UserMessage, AssistantMessage, ToolMessage
from ..types.event import AssistantMessageEvent, StreamMessageGenerator, TextChunkEvent, ToolCallChunkEvent, UsageChunkEvent


class OpenAIProviderMessageParser(BaseMessageParser[
    ChatCompletionChunk,
    ChatCompletion,
    ChatCompletionMessageParam,
]):
    @staticmethod
    def _attachment_to_content_part(attachment: Attachment) -> ChatCompletionContentPartParam:
        match attachment:
            case ImageAttachment() if attachment.source.type == "url":
                return ChatCompletionContentPartImageParam(
                    type="image_url",
                    image_url={"url": attachment.source.url,
                               "detail": "auto"})
            case ImageAttachment() if attachment.source.type == "base64":
                return ChatCompletionContentPartImageParam(
                    type="image_url",
                    image_url={"url": f"data:{attachment.source.mime_type};base64,{attachment.source.data}",
                               "detail": "auto"})
            case AudioAttachment() if attachment.source.type == "base64":
                extname = attachment.source.mime_type.split("/")[-1]
                if extname not in ["mp3", "wav"]:
                    raise AttachmentTypeNotSupportedError(f"audio/{extname}")
                return ChatCompletionContentPartInputAudioParam(
                    type="input_audio",
                    input_audio={"data": attachment.source.data,
                                 "format": cast(Literal["mp3", "wav"], extname)})
            case _:
                raise AttachmentTypeNotSupportedError(attachment.type)

    @override
    @staticmethod
    def normalize_chunk(chunk: ChatCompletionChunk) -> list[TextChunkEvent | ToolCallChunkEvent | UsageChunkEvent] | None:
        if len(chunk.choices) == 0: return None

        result = []

        if chunk.usage:
            result.append(UsageChunkEvent(
                input_tokens=chunk.usage.prompt_tokens,
                output_tokens=chunk.usage.completion_tokens,
                total_tokens=chunk.usage.total_tokens))

        delta: ChoiceDelta = chunk.choices[0].delta
        if delta.content:
            result.append(TextChunkEvent(delta.content))
        if delta.tool_calls:
            for tool_call in delta.tool_calls:
                result.append(ToolCallChunkEvent(
                    tool_call.id,
                    name=tool_call.function and tool_call.function.name,
                    arguments=tool_call.function and tool_call.function.arguments,
                    index=tool_call.index))
        return result

    @override
    @staticmethod
    def to_message(response: ChatCompletion) -> AssistantMessage:
        if len(response.choices) == 0:
            raise ValueError("Empty response")

        usage = response.usage
        choice = response.choices[0]
        message = choice.message

        tool_calls = None
        if message.tool_calls:
            tool_calls = [AssistantMessage.ToolCall(
                id=tool_call.id,
                name=tool_call.function.name,
                arguments=json.loads(tool_call.function.arguments),
            ) for tool_call in message.tool_calls
              if tool_call.type == "function"]

        return AssistantMessage(
            content=message.content,
            tool_calls=tool_calls,
            usage=AssistantMessage.Usage(
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.completion_tokens,
                total_tokens=usage.total_tokens,
            ) if usage else None,
        )

    @override
    @staticmethod
    def from_message(message: ChatMessage) -> ChatCompletionMessageParam:
        match message:
            case SystemMessage():
                return ChatCompletionSystemMessageParam(
                    role=message.role,
                    content=message.content,
                )
            case UserMessage() if message.attachments is None:
                return ChatCompletionUserMessageParam(
                    role=message.role,
                    content=message.content,
                )
            case UserMessage() if message.attachments is not None:
                attachment_contents = [OpenAIProviderMessageParser._attachment_to_content_part(attachment)
                                      for attachment in message.attachments]
                return ChatCompletionUserMessageParam(
                    role=message.role,
                    content=[
                        ChatCompletionContentPartTextParam(text=message.content, type="text"),
                        *attachment_contents
                    ],
                )
            case AssistantMessage():
                message_param = ChatCompletionAssistantMessageParam(
                    role=message.role,
                    content=message.content,
                )
                if message.tool_calls is not None:
                    tool_calls = [ChatCompletionMessageFunctionToolCallParam(
                        type="function",
                        id=tool_call.id,
                        function={
                            "name": tool_call.name,
                            "arguments": json.dumps(tool_call.arguments, ensure_ascii=False),
                        },
                    ) for tool_call in message.tool_calls]
                    message_param["tool_calls"] = tool_calls
                return message_param
            case ToolMessage():
                if message.result is None and message.error is None:
                    raise ValueError(f"ToolMessage({message.id}, {message.name}) is incomplete, "
                                        "result and error cannot be both None")
                if message.error is not None:
                    content = json.dumps({"error": message.error}, ensure_ascii=False)
                else:
                    assert message.result is not None
                    content = message.result

                return ChatCompletionToolMessageParam(
                    role=message.role,
                    content=content,
                    tool_call_id=message.call_id,
                )
            case _:
                raise NotImplementedError

class OpenAIProviderParamParser(BaseParamParser[
    CompletionCreateParamsNonStreaming,
    CompletionCreateParamsStreaming
]):
    def _preparse_tools(self, params: LlmRequestParams) -> list[ChatCompletionFunctionToolParam] | None:
        extracted_tool_likes = params.extract_tools()
        if extracted_tool_likes is None: return None

        tool_schemas = prepare_tools(extracted_tool_likes)
        return [ChatCompletionFunctionToolParam(
            type="function",
            function=cast(FunctionDefinition, tool_schema),
        ) for tool_schema in tool_schemas]

    def _preparse_messages(self, params: LlmRequestParams) -> list[ChatCompletionMessageParam]:
        transformed_messages: list[ChatCompletionMessageParam] = []
        if params.instructions is not None:
            transformed_messages.append(ChatCompletionSystemMessageParam(
                role="system",
                content=params.instructions,
            ))
        for message in params.messages:
            if (type(message) is ToolMessage and not message.is_complete):
                continue
            parsed_message = self._message_parser.from_message(message)
            transformed_messages.append(parsed_message)
        return transformed_messages

    @override
    def parse_nonstream(self, params: LlmRequestParams) -> CompletionCreateParamsNonStreaming:
        result_params = CompletionCreateParamsNonStreaming(
            model=params.model,
            messages=self._preparse_messages(params),
            tool_choice=params.tool_choice,
            temperature=params.temperature,
            max_tokens=params.max_tokens,
            **(params.extra_args or {})
        )
        if (tools := self._preparse_tools(params)) is not None:
            result_params["tools"] = cast(list[ChatCompletionFunctionToolParam], tools)
        return result_params

    @override
    def parse_stream(self, params: LlmRequestParams) -> CompletionCreateParamsStreaming:
        base_params = cast(CompletionCreateParamsStreaming, self.parse_nonstream(params))
        base_params["stream"] = True
        base_params["stream_options"] = {"include_usage": True}
        return base_params

class OpenAIProvider(BaseProvider):
    def __init__(self, base_url: str, api_key: str):
        self._client = AsyncOpenAI(
            base_url=base_url,
            api_key=api_key,
        )
        self._message_parser = OpenAIProviderMessageParser()
        self._param_parser = OpenAIProviderParamParser(self._message_parser)

    @staticmethod
    def _parse_thinking_content(message: AssistantMessage) -> AssistantMessage:
        def match_thinking_content(text: str) -> tuple[str, str | None]:
            pattern = r"<(think|thinking)>(.*?)</\1>"
            match = re.match(pattern, text, re.DOTALL)
            if match:
                start, end = match.span()
                thinking_content = match.group(2)
                remaining_content = text[:start] + text[end:]
                return remaining_content.strip(), thinking_content
            return text, None

        if message.content is None or message.reasoning_content is not None:
            return message.model_copy()
        content, reasoning_content = match_thinking_content(message.content)
        return message.model_copy(
            update={
                "content": content,
                "reasoning_content": reasoning_content,
            })

    @override
    async def list_models(self) -> list[str]:
        models = await self._client.models.list()            
        return [model.id for model in models.data]

    @override
    async def request_nonstream(self, params: LlmRequestParams):
        parsed = self._param_parser.parse_nonstream(params)
        response = await self._client.chat.completions.create(
            **parsed,
            extra_headers=params.headers,
        )
        message = self._message_parser.to_message(response)
        return self._parse_thinking_content(message)

    @override
    async def request_stream(self, params: LlmRequestParams) -> StreamMessageGenerator:
        parsed = self._param_parser.parse_stream(params)
        response = await self._client.chat.completions.create(
            **parsed,
            extra_headers=params.headers,
        )
        message_collector = StreamMessageCollector()
        async for chunk in response:
            normalized_chunks = self._message_parser.normalize_chunk(chunk)
            if normalized_chunks is None: continue
            for normalized in normalized_chunks:
                yield normalized
                message_collector.collect(normalized)

        full_message = message_collector.get_message()
        full_message = self._parse_thinking_content(full_message)
        yield AssistantMessageEvent(message=full_message)



from enum import Enum
from .base_provider import BaseProvider
from .openai import OpenAIProvider

class LlmProviders(str, Enum):
    OPENAI = "openai"

__all__ = [
    "LlmProviders",
    "BaseProvider",
    "OpenAIProvider",
]



import dataclasses
from collections.abc import Callable, Mapping
from types import MethodType
from typing import Any, Awaitable, Literal, TypedDict
from ..logger import logger


class ToolFunctionParameterSchema(TypedDict):
    type: Literal["object"]
    properties: dict[str, Any]
    required: list[str]

class ToolSchema(TypedDict):
    name: str
    description: str
    parameters: ToolFunctionParameterSchema


type ToolFn = Callable[..., Any] | Callable[..., Awaitable[Any]]

"""
RawToolDef example:
{
    "name": "get_current_weather",
    "description": "Get the current weather in a given location",
    "parameters": {
        "type": "object",
        "properties": {
            "location": {
                "type": "string",
                "description": "The city and state, e.g. San Francisco, CA",
            },
            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
        },
        "required": ["location"],
    }
}
"""
type RawToolDef = ToolSchema

@dataclasses.dataclass
class ToolDef:
    name: str
    description: str
    execute: ToolFn
    parameters: ToolFunctionParameterSchema | None = None
    metadata: dict[str, Any] = dataclasses.field(default_factory=dict)
    defaults: Mapping[str, Any] = dataclasses.field(default_factory=dict)

    def executes(self, fn: ToolFn) -> bool:
        """
        Check if this tool's execute function is the same as the given function.

        Handles the case where execute is a bound method by comparing the underlying
        function, so both ``instance.method`` and ``Class.method`` are considered equal.

        Examples:
            ```python
            def some_tool():
                pass

            tool = ToolDef(...) # ToolDef with some_tool as execute
            tool.executes(some_tool) # True
            ```
            ```python
            class BrowserToolset(PythonToolset):
                @python_tool
                def screenshot(self):
                    ...

            tool = BrowserToolset().get_tools()[0] # ToolDef with BrowserToolset.screenshot as execute
            tool.executes(BrowserToolset.screenshot) # True
            tool.executes(BrowserToolset().screenshot) # True
            ```
        """
        def normalize(f: ToolFn) -> ToolFn:
            while hasattr(f, "__wrapped__"):
                f = getattr(f, "__wrapped__")
            if isinstance(f, MethodType):
                return f.__func__
            return f
        return normalize(self.execute) is normalize(fn)

    @staticmethod
    def from_tool_fn(tool_fn: ToolFn) -> "ToolDef":
        if tool_fn.__doc__ is None:
            logger.warning(f"Tool function {tool_fn.__name__} has no docstring, "
                            "which is recommended to be used as the tool description")
        return ToolDef(
            name=tool_fn.__name__,
            description=tool_fn.__doc__ or "",
            execute=tool_fn,
        )

type ToolLike = ToolDef | RawToolDef | ToolFn

__all__ = [
    "ToolFn",
    "ToolDef",
    "RawToolDef",
    "ToolLike",
    "ToolSchema",
]



from abc import ABC, abstractmethod
from ..types.message import AssistantMessage, ChatMessage
from ..types.event import StreamMessageGenerator, TextChunkEvent, ToolCallChunkEvent, UsageChunkEvent
from ..types.request_params import LlmRequestParams


class BaseMessageParser[TChunk, TNonStreamResponse, TProviderMessage](ABC):
    @staticmethod
    @abstractmethod
    def normalize_chunk(chunk: TChunk) -> list[TextChunkEvent | ToolCallChunkEvent | UsageChunkEvent] | None: ...

    @staticmethod
    @abstractmethod
    def to_message(response: TNonStreamResponse) -> AssistantMessage: ...

    @staticmethod
    @abstractmethod
    def from_message(message: ChatMessage) -> TProviderMessage: ...

class BaseParamParser[TNonStreamParams, TStreamParams](ABC):
    def __init__(self, message_parser: BaseMessageParser):
        self._message_parser = message_parser

    @abstractmethod
    def parse_nonstream(self, params: LlmRequestParams) -> TNonStreamParams: ...

    @abstractmethod
    def parse_stream(self, params: LlmRequestParams) -> TStreamParams: ...

class BaseProvider(ABC):
    @abstractmethod
    def __init__(self, base_url: str, api_key: str):
        ...

    @abstractmethod
    async def list_models(self) -> list[str]: ...

    @abstractmethod
    async def request_nonstream(self, params: LlmRequestParams) -> AssistantMessage: ...

    @abstractmethod
    def request_stream(self, params: LlmRequestParams) -> StreamMessageGenerator: ...



import json
import uuid
from abc import ABC
from typing import Any, Literal, Self
from pydantic import BaseModel, ConfigDict, Field, field_validator
from .attachment import Attachment

class ChatMessage(BaseModel, ABC):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))

class SystemMessage(ChatMessage):
    model_config = ConfigDict(json_schema_extra={
        "required": ["content", "role"]
    })

    content: str
    role: Literal["system"] = "system"

class ToolMessage(ChatMessage):
    model_config = ConfigDict(json_schema_extra={
        "required": ["call_id", "name", "arguments", "result", "error", "role", "metadata"]
    })

    call_id: str
    name: str
    arguments: dict[str, Any]
    result: str | None = None
    error: str | None = None
    role: Literal["tool"] = "tool"
    metadata: dict[str, Any] = Field(default_factory=dict)

    @field_validator("result", mode="before")
    def validate_result(cls, v: Any) -> Any:
        if v is None: return v
        if isinstance(v, str): return v
        return json.dumps(v, ensure_ascii=False)

    @property
    def is_complete(self) -> bool:
        return self.result is not None or self.error is not None

    def with_result(self, result: str | None, error: str | None) -> ToolMessage:
        return ToolMessage(
            call_id=self.call_id,
            name=self.name,
            arguments=self.arguments,
            result=result,
            error=error)

class AssistantMessage(ChatMessage):
    class ToolCall(BaseModel):
        id: str
        name: str
        arguments: dict[str, Any]

    class Usage(BaseModel):
        input_tokens: int
        output_tokens: int
        total_tokens: int

        @classmethod
        def default(cls) -> Self:
            return cls(
                input_tokens=0,
                output_tokens=0,
                total_tokens=0)

    model_config = ConfigDict(json_schema_extra={
        "required": ["content", "reasoning_content", "tool_calls", "audio", "images", "usage", "role"]
    })

    content: str | None = None
    reasoning_content: str | None = None
    tool_calls: list[ToolCall] | None = None
    usage: Usage | None = None
    role: Literal["assistant"] = "assistant"

    def get_incomplete_tool_messages(self) -> list[ToolMessage] | None:
        """
        Get a incomplete tool message from the assistant message.
        The returned tool message is incomplete,
        which means it only contains the tool call id, name and arguments.
        Returns None if there is no tool call in the assistant message.
        """
        if self.tool_calls is None: return None
        results: list[ToolMessage] = []
        for tool_call in self.tool_calls:
            results.append(ToolMessage(
                call_id=tool_call.id,
                name=tool_call.name,
                arguments=tool_call.arguments,
                result=None,
                error=None))
        return results

class UserMessage(ChatMessage):
    model_config = ConfigDict(json_schema_extra={
        "required": ["content", "role"]
    })

    content: str
    attachments: list[Attachment] | None = None
    role: Literal["user"] = "user"

__all__ = [
    "ChatMessage",
    "SystemMessage",
    "UserMessage",
    "AssistantMessage",
    "ToolMessage",
]



from contextlib import AsyncExitStack
from typing import Any, override
from mcp import ClientSession, StdioServerParameters as StdioServerParams
from mcp.client.stdio import stdio_client
from .base_mcp_client import McpClient, Tool, ToolResult, McpSessionNotEstablishedError

class LocalServerParams(StdioServerParams): ...

class LocalMcpClient(McpClient):
    def __init__(self, name: str, params: LocalServerParams):
        self._name: str = name
        self._params: LocalServerParams = params
        self._session: ClientSession | None = None
        self._exit_stack: AsyncExitStack | None = None

    @property
    @override
    def name(self) -> str:
        return self._name

    @override
    async def connect(self):
        self._exit_stack = AsyncExitStack()

        try:
            read_stream, write_stream = await self._exit_stack.enter_async_context(
                stdio_client(self._params)
            )
            self._session = await self._exit_stack.enter_async_context(
                ClientSession(read_stream, write_stream)
            )
            await self._session.initialize()
        except BaseException:
            await self.disconnect()
            raise

    @override
    async def list_tools(self) -> list[Tool]:
        if not self._session:
            raise McpSessionNotEstablishedError()

        result = await self._session.list_tools()
        return result.tools

    @override
    async def call_tool(
        self, tool_name: str, arguments: dict[str, Any] | None = None
    ) -> ToolResult:
        if not self._session:
            raise McpSessionNotEstablishedError()

        response = await self._session.call_tool(tool_name, arguments)
        return ToolResult(response.isError, response.content)

    @override
    async def disconnect(self) -> None:
        if self._exit_stack:
            await self._exit_stack.aclose()
            self._session = None
            self._exit_stack = None



from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncGenerator

if TYPE_CHECKING:
    from .message import AssistantMessage

@dataclass(frozen=True)
class TextChunkEvent:
    content: str

@dataclass(frozen=True)
class UsageChunkEvent:
    input_tokens: int
    output_tokens: int
    total_tokens: int

@dataclass(frozen=True)
class ToolCallChunkEvent:
    id: str | None
    name: str | None
    arguments: str | None
    index: int

@dataclass(frozen=True)
class AssistantMessageEvent:
    """
    This event is sent when the assistant message is complete.
    """
    message: AssistantMessage

type StreamMessageEvent = TextChunkEvent | UsageChunkEvent | ToolCallChunkEvent | AssistantMessageEvent
type StreamMessageGenerator = AsyncGenerator[StreamMessageEvent, None]

__all__ = [
    "StreamMessageEvent",
    "StreamMessageGenerator",
    "TextChunkEvent",
    "UsageChunkEvent",
    "ToolCallChunkEvent",
    "AssistantMessageEvent",
]



from abc import ABC, abstractmethod
from typing import Any, NamedTuple
from mcp import Tool as McpTool
from mcp.types import ContentBlock as ToolResultBlock

Tool = McpTool

class ToolResult(NamedTuple):
    is_error: bool
    content: list[ToolResultBlock]

class McpClient(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    async def connect(self): ...
    @abstractmethod
    async def disconnect(self): ...
    @abstractmethod
    async def list_tools(self) -> list[Tool]:
        """
        Raises:
            McpSessionNotEstablishedError: If the session is not established.
        """
    @abstractmethod
    async def call_tool(
        self, tool_name: str, arguments: dict[str, Any] | None = None
    ) -> ToolResult:
        """
        Raises:
            McpSessionNotEstablishedError: If the session is not established.
        """

class McpSessionNotEstablishedError(RuntimeError):
    def __init__(self):
        super().__init__("MCP Session not established, please call connect() first")



from typing import Literal
from pydantic import BaseModel

class Base64Source(BaseModel):
    type: Literal["base64"] = "base64"
    mime_type: str
    data: str

class UrlSource(BaseModel):
    type: Literal["url"] = "url"
    url: str

type FileSource = Base64Source | UrlSource

class ImageAttachment(BaseModel):
    type: Literal["image"] = "image"
    source: FileSource

class DocumentAttachment(BaseModel):
    type: Literal["document"] = "document"
    source: FileSource

class AudioAttachment(BaseModel):
    type: Literal["audio"] = "audio"
    source: FileSource

class VideoAttachment(BaseModel):
    type: Literal["video"] = "video"
    source: FileSource

type Attachment = ImageAttachment | DocumentAttachment | AudioAttachment | VideoAttachment

__all__ = [
    "FileSource",
    "Attachment",

    "Base64Source",
    "UrlSource",
    "ImageAttachment",
    "DocumentAttachment",
    "AudioAttachment",
    "VideoAttachment",
]



import asyncio
import json
from typing import TYPE_CHECKING
from collections.abc import Generator
from .tool.execute import ToolExceptionHandlerManager, execute_tool
from .tool.utils import get_tool_name
from .types import ToolArgumentDecodeError, ToolExecutionError
from .logger import enable_logging

if TYPE_CHECKING:
    from .providers import BaseProvider
    from .types import (
        LlmRequestParams, AssistantMessage, ToolLike,
        StreamMessageEvent, StreamMessageGenerator
    )

class LLM:
    def __init__(self, provider: BaseProvider):
        self._provider = provider
        self._tool_exception_handler_manager = ToolExceptionHandlerManager()

    @property
    def tool_exception_handler_manager(self) -> ToolExceptionHandlerManager:
        return self._tool_exception_handler_manager

    async def execute_tool_call(self,
                                tool: ToolLike,
                                arguments: str | dict) -> tuple[str | None, str | None]:
        """
        Returns:
            A tuple of (result, error)
        """
        result, error = None, None
        try:
            result = await execute_tool(tool, arguments)
        except json.JSONDecodeError as e:
            assert type(arguments) is str
            _error = ToolArgumentDecodeError(get_tool_name(tool), arguments, e)
            error = self._tool_exception_handler_manager.handle(_error)
        except Exception as e:
            _error = ToolExecutionError(tool, arguments, e)
            error = self._tool_exception_handler_manager.handle(_error)
        return result, error

    def execute_tool_call_sync(self,
                               tool: ToolLike,
                               arguments: str | dict
                               ) -> tuple[str | None, str | None]:
        """
        Synchronous wrapper of `execute_tool_call`.
        """
        return asyncio.run(self.execute_tool_call(tool, arguments))


    async def list_models(self) -> list[str]:
        return await self._provider.list_models()

    def list_models_sync(self) -> list[str]:
        return asyncio.run(self.list_models())

    async def generate_text(self, params: LlmRequestParams) -> AssistantMessage:
        return await self._provider.request_nonstream(params)

    def generate_text_sync(self, params: LlmRequestParams) -> AssistantMessage:
        return asyncio.run(self.generate_text(params))

    async def stream_text(self, params: LlmRequestParams) -> StreamMessageGenerator:
        async for chunk in self._provider.request_stream(params):
            yield chunk

    def stream_text_sync(self, params: LlmRequestParams) -> Generator[StreamMessageEvent, None, None]:
        with asyncio.Runner() as runner:
            while True:
                gen = self.stream_text(params)
                try:
                    chunk = runner.run(gen.__anext__())
                    yield chunk
                except StopAsyncIteration:
                    break

__all__ = [
    "LLM",
    "enable_logging",
]
